from functools import cached_property
from typing import Optional
from pydantic import BaseModel, Field, RootModel, AliasChoices
from pydantic import HttpUrl


Redirect = RootModel[str]


def to_key(x: str) -> str:
    return x.title().replace(" ", "_")


class Suggestion(BaseModel):
    page_id: int = Field(validation_alias=AliasChoices("pageid", "page_id"))
    image: Optional[HttpUrl] = None

    key_: Optional[str] = Field(default=None, repr=None)  # type: ignore
    title: str

    description: Optional[str] = None

    @cached_property
    def key(self) -> str:
        if self.key_:
            return self.key_

        return to_key(self.title)


QuickSearchResults = RootModel[list[Suggestion]]



from typing import Optional
from markdownify import markdownify as html2md
from markdown import markdown as md2html
from pyquery import PyQuery as jq

from pydantic import BaseModel, Field


def unwrap(i: int, tag: jq, space: str = "\n\n"):
    contents = jq(tag).html()
    if contents is None:
        jq(tag).remove()
    else:
        jq(tag).replace_with(contents + space)


def remove(i: int, tag: jq):
    jq(tag).replace_with("")


def deh2scrt(i: int, tag: jq):
    if tag.text is None:
        return

    jq(tag).replace_with(f"<b>{jq(tag).text()}HEADEREND</b>")


def rename(i: int, tag: jq, tag_name: str):
    contents = jq(tag).html()
    if contents is None:
        jq(tag).remove()
    else:
        jq(tag).replace_with((f"<{tag_name}>{contents}</{tag_name}>"))


class TgHTML(BaseModel):
    text: str
    output: str | None = None
    markdown: str | None = None

    blocklist: set = {}
    ALLOWED_TAGS: tuple | set = Field(
        default=(
            "b",
            "strong",
            "i",
            "em",
            "code",
            "s",
            "strike",
            "del",
            "u",
            "pre",
            "blockquote",
        ),
        serialization_alias="allowed_tags",
    )

    is_wikipedia: bool = True
    enable_preprocess: bool = True

    def __init__(
        self,
        text: Optional[str] = None,
        html: Optional[str] = None,
        **kwargs,
    ) -> "TgHTML":  # type: ignore
        super(TgHTML, self).__init__(text=text or html, **kwargs)
        self.__post_init__()

    def __post_init__(self):
        # 0. clean html and filter shit
        d = jq(self.text.replace("<cite>", "<cite>\n — "))

        d.find("span").filter(
            lambda i, x: jq(x).attr("style") == "font-style:italic;"
        ).each(lambda i, x: rename(i, x, "i"))
        d.find("span.mw-headline").each(deh2scrt)
        d.find("h2").each(lambda i, x: rename(i, x, "p"))
        d.find("cite").each(lambda i, x: rename(i, x, "i"))
        d.find("strong").each(lambda i, x: rename(i, x, "b"))
        d.find("blockquote blockquote").each(lambda i, x: unwrap(i, x, ""))
        d.find("b").filter(
            lambda i, p: p.text is not None and p.text == ("Избранная статья")
        ).each(remove)
        d.find("p").filter(
            lambda i, p: p.text is not None
            and (
                "Это статья о" in p.text or "Vide etiam paginam discretivam:" in p.text
            )
        ).each(remove)
        d.find("div").filter(
            lambda i, p: p.text is not None
            and (p.text.startswith("Эта статья является избранной."))
        ).each(remove)

        d.find("i").filter(
            lambda i, p: p.text is not None
            and p.text.startswith("Вся обновлённая информация была взята")
        ).each(remove)

        self.bulk_remove(
            d,
            "div.navigation-not-searchable",
            "table",
            "aside",
            ".error",
            ".noprint",
            "audio",
            ".thumb",
            "span.error",
            "span.mw-ext-cite-error",
            "p.hatnote",
            "figure",
            "sup.reference a",
            "span.mw-editsection-bracket",
            *self.blocklist,
        )

        d.find("a").each(lambda i, x: unwrap(i, x, ""))

        source = (
            d.html()
            .replace("<i>", "ITALICRESERVEDSIGN")
            .replace("</i>", "ITALICRESERVEDSIGN")
        )

        # 1. to markdown
        self.markdown = html2md(
            source,
            bullets="■•",
        )

        self.markdown = self.markdown.replace("ITALICRESERVEDSIGN", "_")

        # 2. make some markdown features ignorable
        #    in next step
        # self.md = self.md.replace("\n* ", "\n■ ")

        # 3. to html
        html = md2html(self.markdown)  # type: ignore
        tag = jq(html)

        # 4. remove default html shit like as p and div
        tag.find("div").each(lambda i, x: unwrap(i, x, ""))
        tag.find("p").each(unwrap)

        tag.find("*").filter(lambda i, x: x.tag not in self.ALLOWED_TAGS).each(remove)

        # 5. shitcodded fixes in the end
        self.output = (
            str(tag.html())
            .replace("HEADEREND</strong>\n\n", "</strong>")
            .replace("\n\n", "\n")
            # .replace("\n\n", "\n")
            # .replace("\n", "\n\n")
            .replace("<blockquote>\n", "<blockquote>")
            .replace("\n</blockquote>", "</blockquote>")
        ).strip()

    def bulk_remove(self, d: jq, *selectors):
        for sel in selectors:
            d.find(sel).each(remove)

    def __str__(self) -> str:
        return self.output or ""

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def parsed(self) -> str:
        return self.output or ""

    @property
    def html(self) -> Optional[str]:
        return self.text



from pydantic import BaseModel, Field, RootModel
from typing import Optional

from .image import Image


class SearchResult(BaseModel):
    title: str
    snippet: Optional[str] = None
    size: Optional[int] = None
    page_id: Optional[int] = Field(None, alias="pageid")


class SearchResultWithDescripion(BaseModel):
    page_id: Optional[int] = Field(None, alias="pageid")
    title: str
    index: int
    thumbnail: Optional[Image] = None
    description: Optional[str] = None
    description_source: Optional[str] = Field(None, alias="descriptionsource")


Search = RootModel[list[SearchResult]]
SearchWithDescription = RootModel[list[SearchResultWithDescripion]]


import pytest
from pytest_lazyfixture import lazy_fixture

from params import Params


@pytest.mark.parametrize("params", [lazy_fixture("wikipedia")])
@pytest.mark.asyncio
async def test_get_page_name(params: Params):
    wiki = params.client
    name = await wiki.get_page_name(8000432)

    assert name == "Патрик Тёрнер"


@pytest.mark.parametrize("params", [lazy_fixture("wikipedia")])
@pytest.mark.asyncio
async def test_summary(params: Params):
    wiki = params.client
    await wiki.summary("Патрик Тёрнер")



import pytest
from params import Params


@pytest.mark.asyncio
async def test_image(params: Params):
    wiki = params.client
    image = await wiki.image(params.image_query)

    if params.check_image_res:
        assert image.width > 0
        assert image.height > 0



from pydantic import BaseModel, Field
from .tghtml import TgHTML

from bs4 import BeautifulSoup

from typing import Optional

from .section import Section


class Page(BaseModel):
    title: str
    pageid: Optional[int] = None

    text: Optional[str]
    sections: Optional[list[Section]] = None

    tag_blocklist: Optional[list] = Field(None, repr=False)

    @property
    def parsed(self):
        if (html := TgHTML(self.text, self.tag_blocklist).parsed) == "":
            html = TgHTML(self.text, self.tag_blocklist, enable_preprocess=False).parsed

        return html

    @property
    def fixed(self):
        print("Page.fixed is deprecated, please use .parsed!")
        return self.parsed

    @property
    def section_name(self):
        soup = BeautifulSoup(self.text, "lxml")
        spans = soup.find_all("span")

        for span in spans:
            if span.text != "":
                return span.text

            elif not span["id"].startswith("."):
                return span["id"]



from typing import Any
from .clients import MediaWiki, Fandom, Wikipedia, MediaWikiAbstract
from .models import MediawikiUrl

from httpx import URL


def Wikipya(
    lang: str = "ru",
    base_url: str = "https://{lang}.wikipedia.org/w/api.php",
    prefix: str = "/w",
    params: dict[Any, Any] = {},
) -> MediaWikiAbstract:
    match URL(base_url).host.split(".")[-2]:
        case "wikipedia":
            client = Wikipedia
        case "fandom":
            client = Fandom
        case _:
            client = MediaWiki

    return client(url=MediawikiUrl(base_url, lang, prefix), **params)  # type: ignore



from .page import Page
from .image import Image
from .search import (
    Search,
    SearchResult,
    SearchResultWithDescripion,
    SearchWithDescription,
)
from .opensearch import OpenSearch, OpenSearchResult
from .url import MediawikiUrl
from .summary import Summary
from .section import Section
from .suggestion import QuickSearchResults, Suggestion, Redirect


__all__ = (
    "Page",
    "Image",
    "Search",
    "SearchResult",
    "SearchResultWithDescripion",
    "SearchWithDescription",
    "OpenSearch",
    "OpenSearchResult",
    "MediawikiUrl",
    "Summary",
    "Section",
    "QuickSearchResults",
    "Suggestion",
    "Redirect",
)



from pydantic.dataclasses import dataclass
from pydantic import Field
from typing import Optional


@dataclass
class OpenSearch:
    query: str
    _variants: list[str] = Field(default_factory=list, repr=False)
    descriptions: list[str] = Field(default_factory=list, repr=False)
    links: list[str] = Field(default_factory=list, repr=False)
    results: list["OpenSearchResult"] = Field(default_factory=list)

    def __post_init__(self):
        for i, variant in enumerate(self._variants):
            link = self.links[i] if len(self.links) > 0 else None

            self.results.append(OpenSearchResult(variant, link))


@dataclass
class OpenSearchResult:
    title: str
    link: Optional[str] = None



from ..models import SearchWithDescription, SearchResultWithDescripion
from ..exceptions import NotFound

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def search_with_description(
    self: "MediaWikiAbstract", query: str, limit: int = -1, pithumbsize: int = 200
) -> list[SearchResultWithDescripion]:
    _, json = await self.get(
        generator="prefixsearch",
        prop="pageprops|pageimages|description",
        redirects="",
        ppprop="displaytitle",
        piprop="thumbnail",
        pithumbsize=pithumbsize,
        plimit=limit,
        gpssearch=query,
        gpsnamespace=0,
        gpslimit=limit,
    )

    results_raw = sorted(json["query"]["pages"], key=lambda x: x["index"])
    results = SearchWithDescription.model_validate(results_raw).root

    if len(results) == 0:
        raise NotFound("Search can't find anything on your request")
    else:
        return results



from typing import TYPE_CHECKING

from ..exceptions import NotFound
from ..models import Suggestion


if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def fandom_facade_search(
    self: "MediaWikiAbstract", query: str, limit: int = 1, prop: str = "snippet"
) -> list[Suggestion]:
    try:
        return await self.fandom_search(query, limit)
    except NotFound:
        return await self.legacy_search(query, limit, prop)


__all__ = ("fandom_facade_search",)



from ..models import Image
from ..exceptions import NotFound


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def image(
    self: "MediaWikiAbstract", titles: str, pithumbsize: int = 1000
) -> Image:
    _, json = await self.get(
        titles=titles,
        prop="pageimages",
        pilicense="any",
        pithumbsize=pithumbsize,
        piprop="thumbnail",
    )

    try:
        image = json["query"]["pages"][-1]
        thumb = image["thumbnail"]

        return Image(**thumb)
    except AttributeError as e:
        raise NotFound("Not found image") from e



import contextlib
from ..models import Page
from .lib import query2param


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def page(
    self: "MediaWikiAbstract",
    query: str | int,
    section: int = 0,
    to_section: int = 0,
    prop: str = "text",
) -> Page:
    text = ""
    json = {}

    for i in range(section, to_section + 1):
        _, json = await self.get(
            action="parse",
            section=i,
            prop=prop,
            redirects="true",
            **query2param(query),
        )

        with contextlib.suppress(Exception):
            json["parse"]["text"] = json["parse"]["text"]["*"]

        text += json["parse"]["text"]

    json["parse"]["text"] = text

    page = Page.model_validate(json["parse"])
    page.tag_blocklist = self.tag_blocklist

    return page



from ..exceptions import NotFound
from ..models.suggestion import QuickSearchResults, Suggestion

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def search(
    self: "MediaWikiAbstract", query: str, limit: int = 1, prop: str = "snippet"
) -> list[Suggestion]:
    _, json = await self.get(list="search", srsearch=query, srlimit=limit, srprop=prop)

    results = QuickSearchResults.model_validate(json["query"]["search"]).root

    if len(results) == 0:
        raise NotFound("Search can't find anything on your request")
    else:
        return results



from .fetch_all import fetch_all
from .get_page_name import get_page_name
from .image import image
from .opensearch import opensearch
from .page import page
from .search import search
from .summary import summary
from .sections import sections
from .search_with_description import search_with_description
from .fd_search import fandom_search
from .fa_facade_search import fandom_facade_search


__all__ = (
    "fetch_all",
    "get_page_name",
    "image",
    "opensearch",
    "page",
    "search",
    "summary",
    "sections",
    "search_with_description",
    "fandom_search",
    "fandom_facade_search",
)



from ..models import Summary

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def summary(self: "MediaWikiAbstract", title: str) -> Summary:
    res = await self.get_html(f"{self.url.cleaned}api/rest_v1/page/summary/{title}")

    return Summary.model_validate(res.json())



from .lib import query2param
from ..models import Page


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def sections(self: "MediaWikiAbstract", query: str) -> Page:
    _, json = await self.get(
        action="parse",
        prop="sections",
        redirects="true",
        **query2param(query),
    )

    return Page.model_validate(json["parse"])



from pydantic import BaseModel, Field


class Section(BaseModel):
    title: str = Field(alias="line")
    number: str
    index: int
    level: str
    from_title: str = Field(alias="fromtitle")
    bytes_offset: int = Field(alias="byteoffset")



from typing import Any, Union
from ..models import SearchResult


def query2param(query: Union[str, int, SearchResult]) -> Any:
    if query.__class__ is str:
        return {"page": query}
    elif query.__class__ is int:
        return {"pageid": query}
    elif query.pageid is not None:  # type: ignore
        return {"pageid": query.pageid}  # type: ignore
    else:
        return {"page": query.title}  # type: ignore



"""https://fallout.fandom.com/ru/wikia.php
?controller=UnifiedSearchSuggestions
&method=getSuggestions
&query=%D0%BC%D0%B5%D0%B3%D0%B0%D1%82%D0%BE%D0%BD%D0%BD%D0%B0
&format=json
&scope=internal"""

from ..exceptions import NotFound

from typing import TYPE_CHECKING
from ..models import Suggestion, QuickSearchResults

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def fandom_search(
    self: "MediaWikiAbstract", query: str, limit: int = 1
) -> list[Suggestion]:
    _, json = await self.get(
        url=self.url.url.replace("api.php", "wikia.php"),
        query=query,
        limit=limit,
        # FANDOM SHIT
        controller="UnifiedSearchSuggestions",
        method="getSuggestions",
        scope="internal",
    )

    if len(json["suggestions"]) == 0:
        raise NotFound("Search can't find anything on your request")

    sugs: list[Suggestion] = []

    for name in json["suggestions"]:
        id = json["ids"][name]
        image = json["images"][str(id)]

        sugs.append(Suggestion(page_id=id, title=name, image=image))

    results = QuickSearchResults(sugs[:limit])  # type: ignore

    return results.root



from ..exceptions import NotFound
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def get_page_name(self: "MediaWikiAbstract", id: int) -> str:
    _, json = await self.get(pageids=id)

    try:
        return json["query"]["pages"][-1]["title"]
    except AttributeError:
        raise NotFound(f"Not found page with this id: {id}")



TAG_BLOCKLIST = [
    "ol.references",
    ".reference",
    "link",
    "style",
    "img",
    "aside",
    "table",
    "br",
    "span.mw-ext-cite-error",
    "div.thumbinner",
    "div.gallerytext",
    "div.archwiki-template-meta-related-articles-start",
    "div.tright",
    "div.plainlist",
    "div.gametabs",
    ".noprint",
    "span#w4g_rb_area-1",
    "p.caption",
]


DEFAULT_PARAMS: dict[str, str | int] = {
    "format": "json",
    "action": "query",
    "formatversion": 2,
}



from typing import TYPE_CHECKING
from ..models import Page, Image
from ..exceptions import NotFound

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def fetch_all(
    self: "MediaWikiAbstract", query: str | int, to_section: int = 0
) -> tuple[Page, Image | None, str | None]:
    if isinstance(query, int):
        query = await self.get_page_name(query)

    search = await self.search(query)
    opensearch = await self.opensearch(search[0].title)

    result = opensearch.results[0]
    page = await self.page(result.title, to_section=to_section)

    try:
        image = await self.image(page.title)
    except (NotFound, KeyError):
        image = None

    return page, image, result.link



from params import Params


def test_https(params: Params):
    # Use HTTPS instead of HTTP, please. Ignore with ALLOW_HTTP
    wiki = params.client

    is_https_used = wiki.url.base_url.startswith("https://")
    assert params.allow_http or is_https_used



import traceback

try:
    from .aiowiki import Wikipya
    from .exceptions import NotFound
except ImportError:
    print(traceback.format_exc())


__version__ = "4.0.1.post1"
__all__ = ("Wikipya", "NotFound")



import pytest
from params import Params
from pytest_lazyfixture import lazy_fixture


@pytest.mark.asyncio
async def test_search(params: Params):
    wiki = params.client

    results = await wiki.search(params.search_query, limit=params.search_limit)
    assert len(results) == params.search_limit


@pytest.mark.parametrize("params", [lazy_fixture("wikipedia")])
@pytest.mark.asyncio
async def test_search_with_description(params: Params):
    wiki = params.client

    results = await wiki.search_with_description("анкап", limit=params.search_limit)

    assert results[0].title == "Анархо-капитализм"
    assert len(results) == params.search_limit


@pytest.mark.asyncio
async def test_opensearch(params: Params):
    wiki = params.client

    results = await wiki.opensearch(params.search_query, params.search_limit)
    assert len(results.results) == params.search_limit



import httpx

from typing import Any, Optional
from pydantic import BaseModel, Field

from .models import MediawikiUrl
from .constants import TAG_BLOCKLIST, DEFAULT_PARAMS
from .exceptions import ParseError


HEADERS: dict[str, str] = {
    "User-Agent": "Mozilla/5.0 (iPad; CPU OS 12_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148",
    "Accept-Encoding": "gzip",
}


class BaseClient(BaseModel):
    url: MediawikiUrl
    timeout: Optional[int] = 5

    tag_blocklist: list[str] = Field(repr=False, default=TAG_BLOCKLIST)
    default_params: dict[str, str | int] = Field(repr=False, default=DEFAULT_PARAMS)

    client: Any | None = None

    def model_post_init(self, __context: Any) -> None:
        self.client = httpx.AsyncClient(timeout=self.timeout, headers=HEADERS)

    async def get(
        self, url: Optional[str] = None, **params: Any
    ) -> tuple[httpx.Response, Any]:
        res: httpx.Response = await self.client.get(
            str(url or self.url),
            params={**self.default_params, **params},
            follow_redirects=True,
        )

        json: Any = res.json()

        try:
            if error := json.get("error"):
                raise ParseError(f"{error['code']}: {error['info']}")
        except:  # noqa: E722
            pass

        return res, json

    async def get_html(self, url: Optional[str] = None, **params: Any):
        return await self.client.get(
            str(url or self.url), params=params, follow_redirects=True
        )


class MediaWiki(BaseClient):
    from .methods import (
        fetch_all,
        get_page_name,
        image,
        opensearch,
        page,
        search,
        sections,
        search_with_description,
    )


class Wikipedia(MediaWiki):
    from .methods import summary


class Fandom(MediaWiki):
    from .methods import (
        fandom_search,
        fandom_facade_search as search,
    )


class MediaWikiAbstract(BaseClient):
    from .methods import (
        fetch_all,
        get_page_name,
        image,
        opensearch,
        page,
        search,
        search as legacy_search,
        summary,
        sections,
        search_with_description,
        fandom_search,
    )



from ..models import OpenSearch
from ..exceptions import NotFound

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..clients import MediaWikiAbstract


async def opensearch(
    self: "MediaWikiAbstract", query: str, limit: int = 1
) -> OpenSearch:
    _, json = await self.get(
        action="opensearch",
        search=query,
        limit=limit,
    )

    results = OpenSearch(*json)

    if len(results.results) == 0:
        raise NotFound("OpenSearch can't find anything on your request")

    return results
